package main

import "fmt"
import "math"

/*
Note on Go:
All classes in the same directory are implicitly visible but must be in the same package.
Compile with: go build goo.go infra.go
Or build an run with: go run goo.go infra.go
*/

func main () {
  var QG QueryGraph
  QG.R = []uint{10, 20, 20, 10}
  QG.S = map[uint]float64{}
  QG.SetSel(0, 1, 0.01)
  QG.SetSel(1, 2, 0.5)
  QG.SetSel(2, 3, 0.01)

  res := Goo(QG, false)
  fmt.Println(res)

  // Huffman Tree
  var Huffman QueryGraph
  Huffman.R = []uint{35, 10, 20, 20, 15}
  Huffmanres := Goo(Huffman, true)
  fmt.Println(Huffmanres)
}


func Goo (QG QueryGraph, Huffman bool) string {
  n := len(QG.R)
  Trees := make([]*Tree, n) // n Trees in the beginning, shrinks by one every step
  for i := 0; i < n; i++ {
    Trees[i] = &Tree{float64(QG.R[i]), 1 << uint(i), nil, nil, math.NaN(), nil}
  }

  for k := 0; k < n - 1; k++ { // while |Trees| != 1
    // next two for loops: find T_i, T_j such that join size is minimal
    var mincard float64 = math.Inf(1) // positive infinity
    var minI, minJ int
    for i := 0; i < n - k; i++ {
      for j := i+1; j < n - k; j++ {
        var thiscard float64
        if !Huffman { // Join order problem
          thiscard = Trees[i].Card * Trees[j].Card * QG.GetUnionSelTwoSets(Trees[i].Rels, Trees[j].Rels)
        } else {
          thiscard = Trees[i].Card + Trees[j].Card
        }
        if thiscard < mincard {
          mincard = thiscard
          minI = i
          minJ = j
        }
      }
    }

    // remove Trees minI minJ and add their join tree
    // create jointree, use it to overwrite minI, and swap minJ with last element
    var IjoinJ Tree = Tree{mincard,
                           Trees[minI].Rels | Trees[minJ].Rels,
                           Trees[minI],
                           Trees[minJ],
                           math.NaN(), nil}
    Trees[minI] = &IjoinJ
    Trees[n-k-1], Trees[minJ] = Trees[minJ], Trees[n-k-1]
  }

  return Trees[0].ToString()
}
